%{
/*
 * Hybrid Logic Model Checker 0.9 - HL Formula parser
 * 
 * Copyright (C) 2005 Luigi Dragone
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; version 2.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *   
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Luigi Dragone <luigi@luigidragone.com>
 */

#include <stdio.h>
#include "hl.h"

extern char* yytext;
extern HL_FORMULA_t* theFormula;
extern KRIPKE_STRUCT_t* kripkeStruct;
extern SYMBOL_TABLE_t* propSymTable;
extern SYMBOL_TABLE_t* nominalTable;
extern SYMBOL_TABLE_t* modalityTable;
extern SYMBOL_TABLE_t* wvarTable;

void reportUndefinedModalityError(const char* modality) {
   	fprintf(stderr, "Modality %s is undefined\n", modality); 
}

void reportUndefinedWVarError(const char* wvar) {
   	fprintf(stderr, "World variable %s is undefined\n", wvar); 
}

void reportInvalidIdentifierError(const char* wvar) {
   	fprintf(stderr, "Invalid identifier %s\n", wvar); 
}


void yyerror() {
   	fprintf(stderr, "Syntax error before %s\n", yytext); 
}
%}

%union {
   char *stringconst;
   struct HL_FORMULA_s* exp;
}

%left 'I'
%left '|'
%left '&'
%right '!'

%token <stringconst> tIDENTIFIER 
%type <exp> program exp

%start program

%% 
program: exp
         { theFormula = $1;}
;

exp : tIDENTIFIER
      { 
       	unsigned int id;
        id = SYMBOL_TABLE_getLabelIdentifier(propSymTable, $1);
        if(id != HL_INVALID_LABEL) {
#ifdef DEBUG
			fprintf(stderr, "Propositional symbol %s\n", $1);
#endif
        	$$ = HL_FORMULA_newPropSym(kripkeStruct, id);
        } else {
        	id = SYMBOL_TABLE_getLabelIdentifier(nominalTable, $1);
        	if(id != HL_INVALID_LABEL) {
#ifdef DEBUG
				fprintf(stderr, "Nominal %s\n", $1);
#endif
        		$$ = HL_FORMULA_newNominal(kripkeStruct, id);
        	} else {
	        	id = SYMBOL_TABLE_getLabelIdentifier(wvarTable, $1);
    	    	if(id == HL_INVALID_LABEL) {
        			id = SYMBOL_TABLE_addLabel(wvarTable, $1);
            	}      	
#ifdef DEBUG
				fprintf(stderr, "World variable %s\n", $1);
#endif
//        		id = SYMBOL_TABLE_addLabel(wvarTable, $1);
        		$$ = HL_FORMULA_newWVar(kripkeStruct, id);	
        	}  
        }
      }
    | 'T'
      {
#ifdef DEBUG
		fprintf(stderr, "True\n");
#endif
      	$$ = HL_FORMULA_newTrueConst(kripkeStruct);
      }
    | 'F'
      {
#ifdef DEBUG
		fprintf(stderr, "False\n");
#endif
      	$$ = HL_FORMULA_newFalseConst(kripkeStruct);
      }
    | '!' exp
      {
#ifdef DEBUG
		fprintf(stderr, "Not\n");
#endif
      	$$ = HL_FORMULA_newNeg($2);
      }
    | exp '&' exp
      {
#ifdef DEBUG
		fprintf(stderr, "And\n");
#endif
      	$$ = HL_FORMULA_newAnd($1, $3);
      }
    | exp 'I' exp
      {
#ifdef DEBUG
		fprintf(stderr, "Implies\n");
#endif
      	$$ = HL_FORMULA_newNeg(HL_FORMULA_newAnd($1, HL_FORMULA_newNeg($3)));
      }
    | exp '|' exp
      {
#ifdef DEBUG
		fprintf(stderr, "Or\n");
#endif
      	$$ = HL_FORMULA_newNeg(HL_FORMULA_newAnd(HL_FORMULA_newNeg($1), HL_FORMULA_newNeg($3)));
      }

    | '<' tIDENTIFIER '>' '(' exp ')'
      {
       	unsigned int id;
       	id = SYMBOL_TABLE_getLabelIdentifier(modalityTable, $2);
        if(id != HL_INVALID_LABEL) {
#ifdef DEBUG
			fprintf(stderr, "<%s>\n", $2);
#endif
        	$$ = HL_FORMULA_newDia(id, $5);
       	} else {
	       	reportUndefinedModalityError($2);
		    $$ = NULL;
		}
	   }
    | '[' tIDENTIFIER ']' '(' exp ')'
      {
       	unsigned int id;
       	id = SYMBOL_TABLE_getLabelIdentifier(modalityTable, $2);
        if(id != HL_INVALID_LABEL) {
#ifdef DEBUG
			fprintf(stderr, "[%s]\n", $2);
#endif
        	$$ = HL_FORMULA_newNeg(HL_FORMULA_newDia(id, HL_FORMULA_newNeg($5)));
       	} else {
	       	reportUndefinedModalityError($2);
		    $$ = NULL;
		}
	   }
    | '<' tIDENTIFIER '>' '-' '(' exp ')'
      {
       	unsigned int id;
       	id = SYMBOL_TABLE_getLabelIdentifier(modalityTable, $2);
        if(id != HL_INVALID_LABEL) {
#ifdef DEBUG
			fprintf(stderr, "<%s>-\n", $2);
#endif
        	$$ = HL_FORMULA_newInvDia(id, $6);
       	} else {
       		reportUndefinedModalityError($2);
		    $$ = NULL;
		}
	   }    
    | '[' tIDENTIFIER ']' '-' '(' exp ')'
      {
       	unsigned int id;
       	id = SYMBOL_TABLE_getLabelIdentifier(modalityTable, $2);
        if(id != HL_INVALID_LABEL) {
#ifdef DEBUG
			fprintf(stderr, "[%s]-\n", $2);
#endif
        	$$ = HL_FORMULA_newNeg(HL_FORMULA_newInvDia(id, HL_FORMULA_newNeg($6)));
       	} else {
       		reportUndefinedModalityError($2);
		    $$ = NULL;
		}
	   }    
	| 'E' exp
      {
#ifdef DEBUG
		fprintf(stderr, "Exists\n", $2);
#endif
      	$$ = HL_FORMULA_newExists($2);
      }
	| 'A' exp
      {
#ifdef DEBUG
		fprintf(stderr, "For All\n", $2);
#endif
      	$$ = HL_FORMULA_newNeg(HL_FORMULA_newExists(HL_FORMULA_newNeg($2)));
      }
    | '@' tIDENTIFIER '(' exp ')'
      {
		unsigned int id;
	    id = SYMBOL_TABLE_getLabelIdentifier(nominalTable, $2);
   		if(id != HL_INVALID_LABEL) {
#ifdef DEBUG
			fprintf(stderr, "@ nominal %s\n", $2);
#endif
        	$$ = HL_FORMULA_newAtNom(id, $4);
        } else {
        	id = SYMBOL_TABLE_getLabelIdentifier(wvarTable, $2);
        	if(id == HL_INVALID_LABEL) {
        		id = SYMBOL_TABLE_addLabel(wvarTable, $2);
            }
#ifdef DEBUG
			fprintf(stderr, "@ world variable %s\n", $2);
#endif
       		$$ = HL_FORMULA_newAtWVar(id, $4);
          
        }
      }
    | 'B' tIDENTIFIER '(' exp ')'
      {
 		unsigned int id;
       	id = SYMBOL_TABLE_getLabelIdentifier(wvarTable, $2);
       	if(id != HL_INVALID_LABEL) {
#ifdef DEBUG
			fprintf(stderr, "Bind world variable %s\n", $2);
#endif
       		$$ = HL_FORMULA_newBindWVar(id, $4);
       	} else {
	        reportUndefinedWVarError($2);
            $$ = NULL;
        }
      }
    | '(' exp ')'
      {$$ = $2;}
;

%%
