/*
 * Hybrid Logic Model Checker 0.9 - HL Formula data structures and Model Checking algorithms header file
 * 
 * Copyright (C) 2005 Luigi Dragone
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; version 2.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *   
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Luigi Dragone <luigi@luigidragone.com>
 */

#ifndef __HL_H__
#define __HL_H__

/** To import <code>UINT_MAX</code> */
#include <limits.h>
/** I/O data type (typedefs and functions) */
#include <stdio.h>
/** PBL hash table data type (typedefs and functions) */
#include <pbl.h>

/** Boolean data type */
#define HL_BOOLEAN int
/** Boolean false */
#define HL_FALSE 0
/** Boolean true */
#define HL_TRUE 1
/** Value for denote unassigned world variables */
#define HL_UNASSIGNED UINT_MAX
/** Value for denote invalid symbol label */
#define HL_INVALID_LABEL UINT_MAX

/**
 * Array of symbols.
 * 
 * This data type is employed to represent an indexed collection of symbols (propositional,
 * nominal, variables) expressed as strings. It maps numeric identifiers to corresponding
 * string symbol labels.
 */
typedef struct SYMBOL_ARRAY_s {
	/** indexed collection of symbols */
	const char** entries;
	/** length of the collection */
	size_t length;
} ///
SYMBOL_ARRAY_t;

/**
 * Symbol table.
 * 
 * This structure maps a set of symbol labels (string) to integer identifiers.
 */
typedef struct SYMBOL_TABLE_s {
	/** hash table of pairs &lt;label, identifier&gt; */
	pblHashTable_t* hashTable;
	/** number of labels in the table */
	size_t labelCount;
} ///
SYMBOL_TABLE_t;

/**
 * Create a new symbol table.
 * 
 * @return the new symbol table
 */
SYMBOL_TABLE_t* SYMBOL_TABLE_new();

/**
 * Delete a symbol table and its entries
 * 
 * @param t the table to delete
 */
void SYMBOL_TABLE_delete(SYMBOL_TABLE_t* t);

/**
 * Add a label to the table.
 * 
 * @post a copy of the label string is inserted into the table
 * @param t the table to which add the label
 * @param label the label to add
 * @return the identifier assigned to the label
 */
unsigned int SYMBOL_TABLE_addLabel(SYMBOL_TABLE_t* t, const char* label);

/**
 * Return the number of labels in a symbol table
 * 
 * @param t the table
 * @return the number of label in the table
 */
size_t SYMBOL_TABLE_getLabelCount(SYMBOL_TABLE_t* t);

/**
 * Return the identifier assigned to a label in a symbol table
 * 
 * @param t the symbol table
 * @param label the label to look for
 * @return the identifier assigned to given label or <code>HL_INVALID_LABEL</code> 
 * if it hasn't any
 */
unsigned int SYMBOL_TABLE_getLabelIdentifier(SYMBOL_TABLE_t* t, const char* label);

/**
 * Create a new symbol array from a symbol table
 * 
 * @param t the symbol table from which extract the symbol array
 * @return a symbol array
 */
SYMBOL_ARRAY_t* SYMBOL_ARRAY_new(SYMBOL_TABLE_t* t);

/**
 * Return the label assigned to an identifier
 * 
 * @param a the symbol array
 * @param index the identifier
 * @return the label assigne to the identifier, <code>NULL</code> if any
 */
const char* SYMBOL_ARRAY_getEntry(SYMBOL_ARRAY_t* a, unsigned int index);

/**
 * Return the length of a symbol array
 * 
 * @param a the array
 * @return the length of the array
 */
size_t SYMBOL_ARRAY_getLength(SYMBOL_ARRAY_t* a);

/**
 * Delete a symbol array and its entries
 * 
 * @param a the array to delete
 */
void SYMBOL_ARRAY_delete(SYMBOL_ARRAY_t* a);

/**
 * Kripke structure for HL.
 * 
 * The Kripke structure is expressed using the following conventions:
 * <ul>
 * 	<li>worlds are identified by an integer between <code>0</code> and <code>worldNum - 1</code>;</li>
 * 	<li>modalities are identified by an integer between <code>0</code> and <code>modNum - 1</code>;</li>
 * 	<li>propositional symbols are identified by an integer between <code>0</code> and <code>propSymNum - 1</code>;</li>
 * 	<li>nominals (if any) are identified by an integer between <code>0</code> and <code>nominalNum - 1</code>.</li>
 * </ul>
 */
typedef struct KRIPKE_STRUCT_s {
	/** number of worlds */
	unsigned int worldNum;
	/** number of modalities */
	unsigned int modNum;
	/** number of propositional symbols */
	unsigned int propSymNum;
	/** number of nominals */
	unsigned int nominalNum;
	/** accessibility relations indexed on modality, from world and to world identifiers */
	HL_BOOLEAN ***accessibilityRels;
	/** propositional symbol assignments indexed on propositional symbol and world identifiers*/
	HL_BOOLEAN **propSymAssigns;
	/** nominal assignments expressed as world identifiers indexed on nominal identifiers*/
	unsigned int *nominalAssigns;
} ///
 KRIPKE_STRUCT_t;

/**
 * Create a new Kripke structure.
 * 
 * @param worldNum  the number of worlds in the structure
 * @param modNum the number of modalities
 * @param propSymNum the number of propositional symbols
 * @param nominalNum the number of modalities (at least 0)
 * @return the new Kripke structure
 * 
 * The new KS is empty: no world is accessibile from any other, all propositional
 * symbols are false and all nominals are unassigned.
 */
KRIPKE_STRUCT_t* KRIPKE_STRUCT_new(unsigned int worldNum, unsigned int modNum, unsigned int propSymNum, unsigned int nominalNum);

/**
 * Delete a Kripke structure.
 * 
 * @param ks the Kripke structure to delete
 */
void KRIPKE_STRUCT_delete(KRIPKE_STRUCT_t* ks);

/**
 * Check if a modality identifier is valid in a given KS
 * 
 * @param ks the Kripke structure
 * @param modalityId the modality identifier
 * @param <code>HL_TRUE</code> iff the modality identifier is defined, <code>HL_FALSE</code> otherwise
 */
HL_BOOLEAN KRIPKE_STRUCT_isValidModalityId(KRIPKE_STRUCT_t* ks, unsigned int modalityId);

/**
 * Check if a world identifier is valid in a given KS
 * 
 * @param ks the Kripke structure
 * @param worldId the world identifier
 * @param <code>HL_TRUE</code> iff the world identifier is defined, <code>HL_FALSE</code> otherwise
 */
HL_BOOLEAN KRIPKE_STRUCT_isValidWorldId(KRIPKE_STRUCT_t* ks, unsigned int worldId);

/**
 * Check if a propositional symbol identifier is valid in a given KS
 * 
 * @param ks the Kripke structure
 * @param propSymId the propositional symbol identifier
 * @param <code>HL_TRUE</code> iff the propositional symbol identifier is defined, <code>HL_FALSE</code> otherwise
 */
HL_BOOLEAN KRIPKE_STRUCT_isValidPropSymId(KRIPKE_STRUCT_t* ks, unsigned int propSymId);

/**
 * Check if a nominal identifier is valid in a given KS
 * 
 * @param ks the Kripke structure
 * @param nominalId the nominal identifier
 * @param <code>HL_TRUE</code> iff the nominal identifier is defined, <code>HL_FALSE</code> otherwise
 */
HL_BOOLEAN KRIPKE_STRUCT_isValidNominalId(KRIPKE_STRUCT_t* ks, unsigned int nominalId);

/**
 * Check if a world in the KS is accessible from another given a modality.
 * 
 * @param ks the Kripke structure
 * @param modalityId the identifier of the modality
 * @param fromWorldId the identifier of the source world
 * @param toWorldId the identificer of the target world
 * @return <code>HL_TRUE</code> iff the target world is accessibile from the source one 
 * according to given modality, <code>HL_FALSE</code> otherwise
 */
HL_BOOLEAN KRIPKE_STRUCT_isAccessible(KRIPKE_STRUCT_t* ks, unsigned int modalityId, unsigned int fromWorldId, unsigned int toWorldId);

/**
 * Set the accessibility of a world in a KS from another given a modality
 * 
 * @param ks the Kripke structure
 * @param modalityId the identifier of the modality
 * @param fromWorldId the identifier of the source world
 * @param toWorldId the identificer of the target world
 * @param accessibility <code>HL_TRUE</code> iff the target world is accessibile from the source one 
 * according to given modality, <code>HL_FALSE</code> otherwise
 */
void KRIPKE_STRUCT_setAccessibile(KRIPKE_STRUCT_t* ks, unsigned int modalityId, unsigned int fromWorldId, unsigned int toWorldId, HL_BOOLEAN accessibility);

/**
 * Return the assignment of a propositional symbol in a world of the KS.
 * 
 * @param ks the Kripke structure
 * @param propSymId the identifier of the propositional symbol
 * @param worldId the identifier of the world
 * @return <code>HL_TRUE</code> iff the propositional symbol is true in the
 * given world, <code>HL_FALSE</code> otherwise
 */
HL_BOOLEAN KRIPKE_STRUCT_isPropSymTrue(KRIPKE_STRUCT_t* ks, unsigned int propSymId, unsigned int worldId);

/**
 * Set the truth assignment of a propositional symbol in a world of the KS.
 * 
 * @param ks the Kripke structure
 * @param propSymId the identifier of the propositional symbol
 * @param worldId the identifier of the world
 * @param truthAssign <code>HL_TRUE</code> iff the propositional symbol is true in the
 * given world, <code>HL_FALSE</code> otherwise
 */
void KRIPKE_STRUCT_setPropSym(KRIPKE_STRUCT_t* ks, unsigned int propSymId, unsigned int worldId, HL_BOOLEAN truthAssign);

/**
 * Return the assignment of a nominal in a KS.
 * 
 * @param ks the Kripke structure
 * @param nominalId the identifier of the propositional symbol
 * @return the identifier of the world to which the nominal is assigned, 
 * <code>HL_UNASSIGNED</code> if it is unassigned
 */
unsigned int KRIPKE_STRUCT_getNominalAssign(KRIPKE_STRUCT_t* ks, unsigned int nominalId);

/**
 * Set the assignment of a nominal in a KS.
 * 
 * @param ks the Kripke structure
 * @param nominalId the identifier of the propositional symbol
 * @param worldId the identifier of the world to which the nominal is assigned, 
 * <code>HL_UNASSIGNED</code> if it is unassigned
 */
void KRIPKE_STRUCT_setNominalAssign(KRIPKE_STRUCT_t* ks, unsigned int nominalId, unsigned int worldId);

/**
 * Variable assignment for HL formulas.
 * 
 * The variable assignment is expressed assuming that variables are identified 
 * by an integer between <code>0</code> and <code>wvarNum - 1</code> and they 
 * can assigned to an integer value in the range <code>0..worldNum - 1</code>
 * denoting a model's world or to <code>HL_UNASSIGNED</code> 
 * if they are unassigned.
 */
typedef struct WVAR_ASSIGNMENT_s {
	/** number of distinct world variable symbols */
	unsigned int wvarNum;			
	/** world variable assignments */
	unsigned int* wvarAssigns; 		
} ///
 WVAR_ASSIGNMENT_t;

/**
 * Create a new variable assignement object for a given number of variables.
 * 
 * @param wvarNum the number of variable to assign
 * @return an empty variable assignment object
 */
WVAR_ASSIGNMENT_t* WVAR_ASSIGNMENT_new(unsigned int wvarNum);

/**
 * Delete a variable assignment object
 * 
 * @param g the variable assignment object to delete
 */
void WVAR_ASSIGNMENT_delete(WVAR_ASSIGNMENT_t* g);

/**
 * Create a copy of a variable assignment object.
 * 
 * @param g the object to copy
 * @return a new copy
 */
WVAR_ASSIGNMENT_t* WVAR_ASSIGNMENT_clone(WVAR_ASSIGNMENT_t* g);

/**
 * Return the assignment of a given variable.
 * 
 * @param g the variable assignment object
 * @param wvarId the identifier of the variable
 * @return the identifier of the world to which the variable is assigned,
 * <code>HL_UNASSIGNED</code> if it is unassigned
 */
unsigned int WVAR_ASSIGNMENT_getWVarAssign(WVAR_ASSIGNMENT_t* g, unsigned int wvarId);

/**
 * Set the assignment of a variable.
 * 
 * @param g the variable assignment object
 * @param wvarId the identifier of the variable
 * @param worldId the identifier of the world to which the variable is assigned,
 * <code>HL_UNASSIGNED</code> if it is unassigned
 */
void WVAR_ASSIGNMENT_setWVarAssign(WVAR_ASSIGNMENT_t* g, unsigned int wvarId, unsigned int worldId);

/**
 * Allocate a new assignment object overrinding the specified world variable.
 * 
 * @param g the original assignment object
 * @param wvarId the variable to assign
 * @param worldId the new variable assignment
 * @return the new assignment object
 */
//WVAR_ASSIGNMENT_t* WVAR_ASSIGNMENT_assignOverriding(WVAR_ASSIGNMENT_t* g, unsigned int wvarId, unsigned int worldId);

/**
 * HL formula with model evaluation.
 * 
 * This polimorphic recursive type represents HL formula's expression trees
 * with the corresponding evalutation in a given model (if any).
 */
typedef struct HL_FORMULA_s {
	/** size (complexity) of the formula */
	size_t size;
	/** reference to parent element in formula's tree */
	struct HL_FORMULA_s* parent;
	/** formula's model */
	KRIPKE_STRUCT_t* model;
	/** evalutation of the formula in a model's world */
	HL_BOOLEAN* worldEvals;				
	/** kind of the formula */
	enum {
		/** boolean constant <code>T</code> */
		trueK,
		/** boolean constant <code>F</code> */
		falseK,
		/** formula like: <code>p, q, r, ...</code>*/
		propK, 								
		/** formula like: <code>i, j, k, ...</code> */
		nomK, 								
		/** formula like: <code>x, y, z, ...</code> */
		wvarK, 								
		/** formula like: <code>not(alpha)</code> */
		negK, 								
		/** formula like: <code>alpha and beta</code> */
		andK, 								
		/** formula like: <code>&lt;pi&gt; alpha</code> */
		diaK, 								
		/** formula like: <code>&lt;pi&gt;- alpha</code> */
		invDiaK, 						
		/** formula like: <code>E alpha</code> */
		existsK, 							
		/** formula like: <code>@i alpha</code> */
		atNomK, 							
		/** formula like: <code>@x alpha</code> */
		atWVarK, 							
		/** formula like: <code>^x.alpha</code> */
		bindWVarK							
	} kind; 								
	/** value of the formula (polimorphic) */
  	union {
  		/** identifier of the propositional symbol or nominal or world variable 
  		 * (<code>kind == propK</code> or <code>kind == nomK</code> or 
  		 * <code>kind == wvarK</code>) */
    	unsigned int identifier;
		/** formula like: <code>alpha and beta</code> (<code>kind == andK</code>) */
    	struct {
    		/** reference to left sub expression <var>alpha</var>*/
    		struct HL_FORMULA_s* leftExp; 
    		/** reference to right sub expression <var>beta</var>*/
    		struct HL_FORMULA_s* rightExp;
    	} andE;
		/** formula like <code>not(alpha)</code> or <code>E alpha</code>
		 * (<code>kind == negK</code> or <code>kind == existsK</code>) */
    	struct {
    		/** reference to subexpression <var>alpha</var>*/
    		struct HL_FORMULA_s* exp;
    	} negE, existsE;
		/** formula like <code>&lt;pi&gt; alpha</code> or <code>&lt;pi&gt;- alpha</code>
		 * (<code>kind == diaK</code> or <code>kind == invDiaK</code>) */
    	struct {
    		/** modality's identifier <var>pi</var>*/
    		unsigned int modalityId; 
    		/** reference to subexpression <var>alpha</var>*/
    		struct HL_FORMULA_s* exp;
    	} diaE, invDiaE;
		/** formula like: <code>@i alpha</code> <code>kind == atNomK</code> */
		struct {
			/** nominal's identifier <var>i</var>*/
			unsigned int nominalId; 
    		/** reference to subexpression <var>alpha</var>*/
			struct HL_FORMULA_s* exp;
		} atNomE;
		/** formula like <code>@x alpha</code> or <code>^x.alpha</code> 
		 * (<code>kind == atWVarK</code> or <code>kind == bindWVarK</code>) */
		struct {
			/** world variable's identifier <var>x</var>*/
			unsigned int wvarId; 
    		/** reference to subexpression <var>alpha</var>*/
			struct HL_FORMULA_s* exp;
		} atWVarE, bindWVarE;
  	} val;									
} ///
 HL_FORMULA_t;

/**
 * Create a new formula corresponding to the boolean constant <code>T</code>.
 * 
 * @param model the KS of the formula
 * @return the new HL formula
 */
HL_FORMULA_t* HL_FORMULA_newTrueConst(KRIPKE_STRUCT_t* model);

/**
 * Create a new formula corresponding to the boolean constant <code>F</code>.
 * 
 * @param model the KS of the formula
 * @return the new HL formula
 */
HL_FORMULA_t* HL_FORMULA_newFalseConst(KRIPKE_STRUCT_t* model);

/**
 * Create a new formula corresponding to a propositional symbol evaluation.
 * 
 * @param model the KS of the formula
 * @param propSymId the identifier of the propositional symbol
 * @return the new HL formula, <code>NULL</code> if the given propositional symbol
 * is undefined
 */
HL_FORMULA_t* HL_FORMULA_newPropSym(KRIPKE_STRUCT_t* model, unsigned int propSymId);

/**
 * Create a new formula corresponding to a nominal evaluation.
 * 
 * @param model the KS of the formula
 * @param nominalId the identifier of the nominal
 * @return the new HL formula, <code>NULL</code> if the given nominal is undefined
 */
HL_FORMULA_t* HL_FORMULA_newNominal(KRIPKE_STRUCT_t* model, unsigned int nominalId);

/**
 * Create a new formula corresponding to a variable evaluation.
 * 
 * @param model the KS of the formula
 * @param wvarId the identifier of the variable
 * @return the new HL formula
 */
HL_FORMULA_t* HL_FORMULA_newWVar(KRIPKE_STRUCT_t* model, unsigned int wvarId);

/**
 * Create a negated formula <code>not(alpha)</code>.
 * 
 * @param exp the formula corresponding to <code>alpha</code>
 * @return the new HL formula
 */
HL_FORMULA_t* HL_FORMULA_newNeg(HL_FORMULA_t* exp);

/**
 * Create an existentially quantificated formula <code>E alpha</code>.
 * 
 * @param exp the formula corresponding to <var>alpha</var>
 * @return the new HL formula
 */
HL_FORMULA_t* HL_FORMULA_newExists(HL_FORMULA_t* exp);

/**
 * Create an conjunctive formula <code>alpha and beta</code>.
 * 
 * @param leftExp the formula corresponding to <var>alpha</var>
 * @param rightExp the formula corresponding to <var>beta</var>
 * @return the new HL formula or <code>NULL</code> if given subformulas belong to
 * different KS
 */
HL_FORMULA_t* HL_FORMULA_newAnd(HL_FORMULA_t* leftExp, HL_FORMULA_t* rightExp);

/**
 * Create an formula <code>&lt;pi&gt;alpha</code>.
 * 
 * @param mondalityId the identifier of the modality <var>pi</var>
 * @param exp the formula corresponding to <var>alpha</var>
 * @return the new HL formula or <code>NULL</code> if the modality is undefined
 */
HL_FORMULA_t* HL_FORMULA_newDia(unsigned int modalityId, HL_FORMULA_t* exp);

/**
 * Create an formula <code>&lt;pi&gt;-alpha</code>.
 * 
 * @param mondalityId the identifier of the modality <var>pi</var>
 * @param exp the formula corresponding to <code>alpha</var>
 * @return the new HL formula or <code>NULL</code> if the modality is undefined
 */
HL_FORMULA_t* HL_FORMULA_newInvDia(unsigned int modalityId, HL_FORMULA_t* exp);

/**
 * Create an formula <code>@i alpha</code>.
 * 
 * @param nominalId the identifier of the nominal <var>i</var>
 * @param exp the formula corresponding to <code>alpha</code>
 * @return the new HL formula or <code>NULL</code> if the nominal is undefined
 */
HL_FORMULA_t* HL_FORMULA_newAtNom(unsigned int nominalId, HL_FORMULA_t* exp);

/**
 * Create an formula <code>@x alpha</code>.
 * 
 * @param wvarId the identifier of the variable <var>x</var>
 * @param exp the formula corresponding to <var>alpha</var>
 * @return the new HL formula
 */
HL_FORMULA_t* HL_FORMULA_newAtWVar(unsigned int wvarId, HL_FORMULA_t* exp);

/**
 * Create an formula <code>^x alpha</code>.
 * 
 * @param wvarId the identifier of the variable <var>x</var>
 * @param exp the formula corresponding to <var>alpha</var>
 * @return the new HL formula
 */
HL_FORMULA_t* HL_FORMULA_newBindWVar(unsigned int wvarId, HL_FORMULA_t* exp);

/**
 * Delete a HL formula object
 * 
 * @param exp the formula object to delete
 */
void HL_FORMULA_delete(HL_FORMULA_t* exp);

/**
 * Execute the model checking procedure on the given formula.
 * 
 * @param the formula 
 * @param useMCLite <code>HL_TRUE</code> to use <code>MCLite</code>
 * model checking algorithm, otherwise <code>MCFull</code> is employed
 */
void HL_FORMULA_modelChecking(HL_FORMULA_t* exp, HL_BOOLEAN useMCLite);

/**
 * Return the evaluation of the formula in worlds of the KS.
 * 
 * @pre the model checking procedure should be completed before
 * @param exp the formula
 * @return a vector of <code>worldNum</code> items containing the evalutation
 * of the formula in the corresponding KS's world
 */
HL_BOOLEAN* HL_FORMULA_getWorldEvals(HL_FORMULA_t* exp);

/**
 * Print the formula to an output stream.
 * 
 * @param os the output strem
 * @param exp the formula to print
 * @param propSymLabels the symbol array of propositional symbols occurring in the formula
 * @param nominalLabels the symbol array of nominals occurring in the formula
 * @param modalityLabels the symbol array of modalitiess occurring in the formula
 * @param wvarLabels the symbol array of world variables occurring in the formula
 */ 
void HL_FORMULA_prettyPrint(FILE* os, HL_FORMULA_t* exp, SYMBOL_ARRAY_t* propSymLabels, 
	SYMBOL_ARRAY_t* nominalLabels, SYMBOL_ARRAY_t* modalityLabels, SYMBOL_ARRAY_t* wvarLabels);


/**
 * Print the model checking procedure statistics.
 * 
 * @pre the model checking procedure should be completed before
 * @param os the output stream
 * @param useMCLite <code>HL_TRUE</code> if the <code>MCLite</code> procedure has been executed
 */
void HL_FORMULA_printMCStats(FILE* os, HL_BOOLEAN useMCLite);

#endif
